/*
 * AdvancedPhotoPanel.java 5 mai 2009
 *
 * Copyright (c) 2009 Emmanuel PUYBARET / eTeks <info@eteks.com>. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.eteks.sweethome3d.plugin.advancedrendering;

import java.awt.BasicStroke;
import java.awt.CardLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.EventQueue;
import java.awt.FlowLayout;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.RenderingHints;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.geom.Ellipse2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.Line2D;
import java.awt.image.BufferedImage;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.IOException;
import java.lang.ref.WeakReference;
import java.text.DateFormat;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.ResourceBundle;
import java.util.TimeZone;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import javax.imageio.ImageIO;
import javax.swing.ActionMap;
import javax.swing.ButtonGroup;
import javax.swing.DefaultListCellRenderer;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFormattedTextField;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JSeparator;
import javax.swing.JSpinner;
import javax.swing.KeyStroke;
import javax.swing.SpinnerDateModel;
import javax.swing.SpinnerNumberModel;
import javax.swing.SwingUtilities;
import javax.swing.JSpinner.DefaultEditor;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.text.DefaultFormatterFactory;
import javax.swing.text.NumberFormatter;

import com.eteks.sweethome3d.model.AspectRatio;
import com.eteks.sweethome3d.model.Home;
import com.eteks.sweethome3d.model.HomeEnvironment;
import com.eteks.sweethome3d.model.UserPreferences;
import com.eteks.sweethome3d.swing.AutoCommitSpinner;
import com.eteks.sweethome3d.swing.FileContentManager;
import com.eteks.sweethome3d.swing.HomeComponent3D;
import com.eteks.sweethome3d.swing.IconManager;
import com.eteks.sweethome3d.swing.PhotoPanel;
import com.eteks.sweethome3d.swing.ResourceAction;
import com.eteks.sweethome3d.swing.ScaledImageComponent;
import com.eteks.sweethome3d.swing.SwingTools;
import com.eteks.sweethome3d.tools.OperatingSystem;
import com.eteks.sweethome3d.viewcontroller.View;

/**
 * A panel to edit photo creation with advanced parameters.
 * Based on com.eteks.sweethome3d.swing.PhotoPanel class in Sweet Home 3D 2.3.
 * @author Emmanuel Puybaret
 */
public class AdvancedPhotoPanel extends JPanel {
  private enum ActionType {START_PHOTO_CREATION, STOP_PHOTO_CREATION, SAVE_PHOTO, CLOSE}

  private static final String VISUAL_PROPERTY_PREFIX = "com.eteks.sweethome3d.plugin.advancedrendering.AdvancedPhotoPanel.";
  private static final String RATIO_2_1 = "2 / 1";
  
  private static final String WAIT_CARD  = "wait";
  private static final String PHOTO_CARD = "photo";
  
  private final Home                 home;
  private final UserPreferences      preferences;
  private ScaledImageComponent       photoComponent;
  private JLabel                     animatedWaitLabel;
  private JLabel                     widthLabel;
  private JSpinner                   widthSpinner;
  private JLabel                     heightLabel;
  private JSpinner                   heightSpinner;
  private JLabel                     lensLabel;
  private JComboBox                  lensComboBox;
  private JCheckBox                  applyProportionsCheckBox;
  private JComboBox                  aspectRatioComboBox;
  private JLabel                     qualityLabel;
  private JRadioButton               fastQualityRadioButton;
  private JRadioButton               bestQualityRadioButton;
  private JLabel                     dateLabel;
  private JSpinner                   dateSpinner;
  private JLabel                     timeLabel;
  private JSpinner                   timeSpinner;
  private JLabel                     timeZoneLabel;
  private JComboBox                  timeZoneComboBox;
  private JLabel                     northDirectionLabel;
  private JSpinner                   northDirectionSpinner;
  private JComponent                 northDirectionComponent;
  private JLabel                     longitudeLabel;
  private JSpinner                   longitudeSpinner;
  private JLabel                     latitudeLabel;
  private JSpinner                   latitudeSpinner;
  private JCheckBox                  useCeilingLightCheckBox;
  private JCheckBox                  useFurnitureLightCheckBox;
  private JCheckBox                  addSunlightCheckBox;
  private JCheckBox                  shinyWallsCheckBox;
  private JCheckBox                  shinyFloorsCheckBox;
  private JCheckBox                  shinyCeilingsCheckBox;

  private JPanel                     photoPanel;
  private CardLayout                 photoCardLayout;
  private ExecutorService            photoCreationExecutor;
  private JButton                    createButton;
  private JButton                    saveButton;
  private JButton                    closeButton;
  private String                     dialogTitle;

  private ChangeListener             widthChangeListener;
  private ChangeListener             heightChangeListener;

  private int                        photoWidth;
  private int                        photoHeight;
  private Object                     photoAspectRatio;
  private float                      view3DAspectRatio;
  private AdvancedPhotoRenderer.Lens lens;
  private int                        photoQuality;
  private Calendar                   photoCalendar;
  private float                      northDirection;
  private float                      longitude;
  private float                      latitude;
  private boolean                    useCeilingLight;
  private boolean                    useFurnitureLight;
  private boolean                    addSunlight;
  private float                      wallShininess;
  private float                      floorShininess;
  private float                      ceilingShininess;
  
  private static AdvancedPhotoPanel     currentPhotoPanel; // There can be only one photo panel opened at a time
  
  public AdvancedPhotoPanel(Home home, 
                            UserPreferences preferences, 
                            ResourceBundle  additionalResource) {
    super(new GridBagLayout());
    this.home = home;
    this.preferences = preferences;
    
    HomeEnvironment homeEnvironment = home.getEnvironment();
    // Retrieve existing values
    Integer photoWidth = (Integer)home.getVisualProperty(VISUAL_PROPERTY_PREFIX + "photoWidth");
    if (photoWidth == null) {
      this.photoWidth = homeEnvironment.getPhotoWidth();
    } else {
      this.photoWidth = photoWidth.intValue();
    }
    Integer photoHeight = (Integer)home.getVisualProperty(VISUAL_PROPERTY_PREFIX + "photoHeight");
    if (photoHeight == null) {
      this.photoHeight = homeEnvironment.getPhotoHeight();
    } else {
      this.photoHeight = photoHeight.intValue();
    }
    String lens = (String)home.getVisualProperty(VISUAL_PROPERTY_PREFIX + "lens");
    if (lens != null) {
      this.lens = AdvancedPhotoRenderer.Lens.valueOf(lens);
    } else {
      Boolean sphericalCamera = (Boolean)home.getVisualProperty(VISUAL_PROPERTY_PREFIX + "sphericalCamera");
      if (sphericalCamera != null) {
        this.lens = AdvancedPhotoRenderer.Lens.SPHERICAL;
      } else {
        this.lens = AdvancedPhotoRenderer.Lens.PINHOLE;
      }
    }
    String aspectRatioText = (String)home.getVisualProperty(VISUAL_PROPERTY_PREFIX + "photoAspectRatio");
    if (aspectRatioText == null) {
      this.photoAspectRatio = homeEnvironment.getPhotoAspectRatio();
    } else {
      if (RATIO_2_1.equals(aspectRatioText)) {
        this.photoAspectRatio = RATIO_2_1;
      } else {
        this.photoAspectRatio = AspectRatio.valueOf(aspectRatioText);
      }
    }
    Integer photoQuality = (Integer)home.getVisualProperty(VISUAL_PROPERTY_PREFIX + "photoQuality");
    if (photoQuality == null) {
      this.photoQuality = Math.max(2, homeEnvironment.getPhotoQuality());
    } else {
      this.photoQuality = photoQuality.intValue();
    }
    this.photoCalendar = (Calendar)home.getVisualProperty(VISUAL_PROPERTY_PREFIX + "photoCalendar");
    if (this.photoCalendar == null) {
      this.photoCalendar = new GregorianCalendar();
    } 
    Float northDirection = (Float)home.getVisualProperty(VISUAL_PROPERTY_PREFIX + "northDirection");
    if (northDirection != null) {
      this.northDirection = northDirection.floatValue();
    }
    Float latitude = (Float)home.getVisualProperty(VISUAL_PROPERTY_PREFIX + "latitude");
    if (latitude != null) {
      this.latitude = latitude.floatValue();
    } else {
      this.latitude = (float)Math.toRadians(48.83);
    }
    Float longitude = (Float)home.getVisualProperty(VISUAL_PROPERTY_PREFIX + "longitude");
    if (longitude != null) {
      this.longitude = longitude.floatValue();
    } else {
      this.longitude = (float)Math.toRadians(2.303);
    }
    Boolean useCeilingLight = (Boolean)home.getVisualProperty(VISUAL_PROPERTY_PREFIX + "useCeilingLight");
    if (useCeilingLight != null) {
      this.useCeilingLight = useCeilingLight.booleanValue();
    } else {
      this.useCeilingLight = true;
    }
    Boolean useFurnitureLight = (Boolean)home.getVisualProperty(VISUAL_PROPERTY_PREFIX + "useFurnitureLight");
    if (useFurnitureLight != null) {
      this.useFurnitureLight = useFurnitureLight.booleanValue();
    }
    Boolean addSunlight = (Boolean)home.getVisualProperty(VISUAL_PROPERTY_PREFIX + "addSunLight");
    if (addSunlight != null) {
      this.addSunlight = addSunlight.booleanValue();
    }
    Float wallShininess = (Float)home.getVisualProperty(VISUAL_PROPERTY_PREFIX + "wallShininess");
    if (wallShininess != null) {
      this.wallShininess = wallShininess.floatValue();
    }
    Float floorShininess = (Float)home.getVisualProperty(VISUAL_PROPERTY_PREFIX + "floorShininess");
    if (floorShininess != null) {
      this.floorShininess = floorShininess.floatValue();
    }
    Float ceilingShininess = (Float)home.getVisualProperty(VISUAL_PROPERTY_PREFIX + "ceilingShininess");
    if (ceilingShininess != null) {
      this.ceilingShininess = ceilingShininess.floatValue();
    }
    
    createActions(preferences);
    createComponents(preferences, additionalResource);
    setMnemonics(preferences, additionalResource);
    layoutComponents();    

    preferences.addPropertyChangeListener(UserPreferences.Property.LANGUAGE, new LanguageChangeListener(this, additionalResource));
  }
  
  /**
   * Creates actions for variables.
   */
  private void createActions(UserPreferences preferences) {
    final ActionMap actions = getActionMap();
    actions.put(ActionType.START_PHOTO_CREATION, 
        new ResourceAction(preferences, PhotoPanel.class, ActionType.START_PHOTO_CREATION.name(), true) {
          @Override
          public void actionPerformed(ActionEvent ev) {
            startPhotoCreation();
          }
        });
    actions.put(ActionType.STOP_PHOTO_CREATION, 
        new ResourceAction(preferences, PhotoPanel.class, ActionType.STOP_PHOTO_CREATION.name(), true) {
          @Override
          public void actionPerformed(ActionEvent ev) {
            stopPhotoCreation();
          }
        });
    actions.put(ActionType.SAVE_PHOTO, 
        new ResourceAction(preferences, PhotoPanel.class, ActionType.SAVE_PHOTO.name(), false) {
          @Override
          public void actionPerformed(ActionEvent ev) {
            savePhoto();
          }
        });
    actions.put(ActionType.CLOSE, 
        new ResourceAction(preferences, PhotoPanel.class, ActionType.CLOSE.name(), true) {
          @Override
          public void actionPerformed(ActionEvent ev) {
            close();
          }
        });
  }

  /**
   * Creates and initializes components.
   */
  private void createComponents(final UserPreferences preferences, 
                                final ResourceBundle additionalResource) {
    this.photoComponent = new ScaledImageComponent();
    this.photoComponent.setPreferredSize(new Dimension(400, 400));

    this.animatedWaitLabel = new JLabel(new ImageIcon(PhotoPanel.class.getResource("resources/animatedWait.gif")));

    this.widthChangeListener = new ChangeListener() {
        public void stateChanged(ChangeEvent ev) {
          photoWidth = ((Number)widthSpinner.getValue()).intValue();
          // Save value entered by user in a visual property
          home.setVisualProperty(VISUAL_PROPERTY_PREFIX + "photoWidth", photoWidth);
          updatePhotoHeight();
        }
      };
    this.heightChangeListener = new ChangeListener() {
        public void stateChanged(ChangeEvent ev) {
          photoHeight = ((Number)heightSpinner.getValue()).intValue();
          // Save value entered by user in a visual property
          home.setVisualProperty(VISUAL_PROPERTY_PREFIX + "photoHeight", photoHeight);
          
          if (photoAspectRatio != AspectRatio.FREE_RATIO) {
            widthSpinner.getModel().removeChangeListener(widthChangeListener);
            if (photoAspectRatio == AspectRatio.VIEW_3D_RATIO) {
              photoWidth = Math.round(photoHeight * view3DAspectRatio);
            } else if (photoAspectRatio == RATIO_2_1) {
              photoWidth = Math.round(photoHeight * 2);
            } else if (((AspectRatio)photoAspectRatio).getValue() != null) {
              photoWidth = Math.round(photoHeight * ((AspectRatio)photoAspectRatio).getValue());
            }
            widthSpinner.setValue(photoWidth);
            home.setVisualProperty(VISUAL_PROPERTY_PREFIX + "photoWidth", photoWidth);          
            widthSpinner.getModel().addChangeListener(widthChangeListener);
          }
        }
      };

    // Create width label and spinner 
    this.widthLabel = new JLabel();
    final SpinnerNumberModel widthSpinnerModel = new SpinnerNumberModel(480, 10, 3000, 10);
    this.widthSpinner = new AutoCommitSpinner(widthSpinnerModel);
    widthSpinnerModel.setValue(this.photoWidth);
    widthSpinnerModel.addChangeListener(widthChangeListener);
    
    // Create height label and spinner 
    this.heightLabel = new JLabel();
    final SpinnerNumberModel heightSpinnerModel = new SpinnerNumberModel(480, 10, 3000, 10);
    this.heightSpinner = new AutoCommitSpinner(heightSpinnerModel);
    heightSpinnerModel.setValue(this.photoHeight);
    heightSpinnerModel.addChangeListener(heightChangeListener);

    // Create lens label and combo box
    this.lensLabel = new JLabel();
    this.lensComboBox = new JComboBox(AdvancedPhotoRenderer.Lens.values());
    this.lensComboBox.setRenderer(new DefaultListCellRenderer() {
        @Override
        public Component getListCellRendererComponent(JList list, Object value, 
                                                      int index, boolean isSelected, boolean cellHasFocus) {
          String displayedValue = "";
          switch ((AdvancedPhotoRenderer.Lens)value) {
            case SPHERICAL :
              displayedValue = additionalResource.getString("AdvancedPhotoPanel.lensComboBox.sphericalLens.text");
              break;
            case FISHEYE :
              displayedValue = additionalResource.getString("AdvancedPhotoPanel.lensComboBox.fisheyeLens.text");
              break;
            case DEPTH_OF_FIELD :
              displayedValue = additionalResource.getString("AdvancedPhotoPanel.lensComboBox.depthOfFieldLens.text");
              break;
            case PINHOLE :
            default :
              displayedValue = additionalResource.getString("AdvancedPhotoPanel.lensComboBox.pinholeLens.text");
              break;
          }
          return super.getListCellRendererComponent(list, displayedValue, index, isSelected,
              cellHasFocus);
        }
      });
    this.lensComboBox.setSelectedItem(this.lens);
    this.lensComboBox.addItemListener(new ItemListener() {
        public void itemStateChanged(ItemEvent ev) {
          lens = (AdvancedPhotoRenderer.Lens)lensComboBox.getSelectedItem();
          home.setVisualProperty(VISUAL_PROPERTY_PREFIX + "lens", lens.name()); 
          if (lens == AdvancedPhotoRenderer.Lens.SPHERICAL) {
            applyProportionsCheckBox.setSelected(true);
            aspectRatioComboBox.setSelectedItem(RATIO_2_1);
          } else if (lens == AdvancedPhotoRenderer.Lens.FISHEYE) {
            applyProportionsCheckBox.setSelected(true);
            aspectRatioComboBox.setSelectedItem(AspectRatio.SQUARE_RATIO);
          }  
          updatePhotoHeight();
        }
      });

    // Create apply proportions check box 
    boolean notFreeAspectRatio = this.photoAspectRatio != AspectRatio.FREE_RATIO;
    this.applyProportionsCheckBox = new JCheckBox();
    this.applyProportionsCheckBox.setSelected(notFreeAspectRatio);
    final ItemListener applyProportionsListener = new ItemListener() {
        public void itemStateChanged(ItemEvent ev) {
          photoAspectRatio = applyProportionsCheckBox.isSelected()
              ? aspectRatioComboBox.getSelectedItem()
              : AspectRatio.FREE_RATIO;
          // Save value entered by user in a visual property
          home.setVisualProperty(VISUAL_PROPERTY_PREFIX + "photoAspectRatio", 
              photoAspectRatio instanceof AspectRatio 
                ? ((AspectRatio)photoAspectRatio).name()
                : photoAspectRatio);
              
          updatePhotoHeight();
        }
      };
    this.applyProportionsCheckBox.addItemListener(applyProportionsListener);
    this.aspectRatioComboBox = new JComboBox(new Object [] {
        AspectRatio.VIEW_3D_RATIO,
        AspectRatio.SQUARE_RATIO,
        AspectRatio.RATIO_4_3,
        AspectRatio.RATIO_3_2,
        AspectRatio.RATIO_16_9,
        RATIO_2_1});
    this.aspectRatioComboBox.setRenderer(new DefaultListCellRenderer() {
        @Override
        public Component getListCellRendererComponent(JList list, Object value, 
                                                      int index, boolean isSelected, boolean cellHasFocus) {
          String displayedValue = "";
          if (RATIO_2_1.equals(value)) {
            // AspectRatio enum can't be extended so use a string for ratio 2/1
            displayedValue = RATIO_2_1;
          } else {
            AspectRatio aspectRatio = (AspectRatio)value;
            if (aspectRatio != AspectRatio.FREE_RATIO) {
              switch (aspectRatio) {
                case VIEW_3D_RATIO :
                  displayedValue = preferences.getLocalizedString(
                      PhotoPanel.class, "aspectRatioComboBox.view3DRatio.text");
                  break;
                case SQUARE_RATIO :
                  displayedValue = preferences.getLocalizedString(
                      PhotoPanel.class, "aspectRatioComboBox.squareRatio.text");
                  break;
                case RATIO_4_3 :
                  displayedValue = "4/3";
                  break;
                case RATIO_3_2 :
                  displayedValue = "3/2";
                  break;
                case RATIO_16_9 :
                  displayedValue = "16/9";
                  break;
              }
            }
          }
          return super.getListCellRendererComponent(list, displayedValue, index, isSelected,
              cellHasFocus);
        }
      });
    this.aspectRatioComboBox.addItemListener(new ItemListener() {
        public void itemStateChanged(ItemEvent ev) {
          applyProportionsListener.itemStateChanged(null);
        }
      });
    this.aspectRatioComboBox.setEnabled(notFreeAspectRatio);
    this.aspectRatioComboBox.setSelectedItem(this.photoAspectRatio);

    // Quality label and radio buttons 
    this.qualityLabel = new JLabel();
    this.fastQualityRadioButton = new JRadioButton();
    this.fastQualityRadioButton.setSelected(this.photoQuality == 2);
    this.bestQualityRadioButton = new JRadioButton();
    this.bestQualityRadioButton.setSelected(this.photoQuality == 3);
    ItemListener qualityChangeListener = new ItemListener() {
        public void itemStateChanged(ItemEvent ev) {
          if (fastQualityRadioButton.isSelected()) {
            photoQuality = 2;
          } else {
            photoQuality = 3;
          }
          // Save value entered by user in a visual property
          home.setVisualProperty(VISUAL_PROPERTY_PREFIX + "photoQuality", photoQuality);
        }
      };
    this.fastQualityRadioButton.addItemListener(qualityChangeListener);
    this.bestQualityRadioButton.addItemListener(qualityChangeListener);
    ButtonGroup group = new ButtonGroup();
    group.add(this.fastQualityRadioButton);
    group.add(bestQualityRadioButton);

    
    this.dateLabel = new JLabel();
    final SpinnerDateModel dateSpinnerModel = new SpinnerDateModel();
    dateSpinnerModel.setValue(this.photoCalendar.getTime());
    this.dateSpinner = new JSpinner(dateSpinnerModel);
    JSpinner.DateEditor dateEditor = new JSpinner.DateEditor(this.dateSpinner, 
        ((SimpleDateFormat)DateFormat.getDateInstance(DateFormat.SHORT)).toPattern());
    this.dateSpinner.setEditor(dateEditor);
    SwingTools.addAutoSelectionOnFocusGain(dateEditor.getTextField());
    
    this.timeLabel = new JLabel();
    final SpinnerDateModel timeSpinnerModel = new SpinnerDateModel();
    timeSpinnerModel.setValue(this.photoCalendar.getTime());
    this.timeSpinner = new JSpinner(timeSpinnerModel);
    JSpinner.DateEditor timeEditor = new JSpinner.DateEditor(this.timeSpinner, 
        ((SimpleDateFormat)DateFormat.getTimeInstance(DateFormat.SHORT)).toPattern());
    this.timeSpinner.setEditor(timeEditor);
    SwingTools.addAutoSelectionOnFocusGain(timeEditor.getTextField());

    this.timeZoneLabel = new JLabel();
    String [] timeZoneIDs = TimeZone.getAvailableIDs();
    Arrays.sort(timeZoneIDs);
    this.timeZoneComboBox = new JComboBox(timeZoneIDs);
    this.timeZoneComboBox.setSelectedItem(this.photoCalendar.getTimeZone().getID());
    this.timeZoneComboBox.setRenderer(new DefaultListCellRenderer() {
        @Override
        public Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected,
                                                      boolean cellHasFocus) {
          return super.getListCellRendererComponent(list, value, index, isSelected, cellHasFocus);
        }
      });

    final ChangeListener dateTimeChangeListener = new ChangeListener() {
        public void stateChanged(ChangeEvent ev) {
          // Merge date and time
          GregorianCalendar dateCalendar = new GregorianCalendar();
          dateCalendar.setTime((Date)dateSpinnerModel.getValue());
          GregorianCalendar timeCalendar = new GregorianCalendar();
          timeCalendar.setTime((Date)timeSpinnerModel.getValue());
          dateCalendar.set(GregorianCalendar.HOUR_OF_DAY, timeCalendar.get(GregorianCalendar.HOUR_OF_DAY));
          dateCalendar.set(GregorianCalendar.MINUTE, timeCalendar.get(GregorianCalendar.MINUTE));
          dateCalendar.set(GregorianCalendar.SECOND, 0);
          dateCalendar.setTimeZone(TimeZone.getTimeZone((String)timeZoneComboBox.getSelectedItem()));
          
          photoCalendar = dateCalendar;
          home.setVisualProperty(VISUAL_PROPERTY_PREFIX + "photoCalendar", photoCalendar);
        }
      };
    dateSpinnerModel.addChangeListener(dateTimeChangeListener);
    timeSpinnerModel.addChangeListener(dateTimeChangeListener);
    this.timeZoneComboBox.addItemListener(new ItemListener() {
        public void itemStateChanged(ItemEvent ev) {
          dateTimeChangeListener.stateChanged(null);
        }
      });
    
    this.northDirectionLabel = new JLabel();
    final SpinnerNumberModel northDirectionSpinnerModel = new SpinnerNumberModel(0, 0, 360, 10);
    this.northDirectionSpinner = new AutoCommitSpinner(northDirectionSpinnerModel);
    northDirectionSpinnerModel.setValue((int)Math.round(Math.toDegrees(this.northDirection)));
    this.northDirectionComponent = new JComponent() {
        @Override
        public Dimension getPreferredSize() {
          return new Dimension(32, 32);
        }
        
        @Override
        protected void paintComponent(Graphics g) {
          Graphics2D g2D = (Graphics2D) g;
          g2D.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
          g2D.translate(getWidth() / 2, getHeight() / 2);
          g2D.scale(getWidth() / 2, getWidth() / 2);
          g2D.rotate(northDirection);
          // Draw a round arc
          g2D.setStroke(new BasicStroke(0.5f / getWidth()));
          g2D.draw(new Ellipse2D.Float(-0.7f, -0.7f, 1.4f, 1.4f));
          g2D.draw(new Line2D.Float(-0.85f, 0, -0.7f, 0));
          g2D.draw(new Line2D.Float(0.85f, 0, 0.7f, 0));
          g2D.draw(new Line2D.Float(0, -0.8f, 0, -0.7f));
          g2D.draw(new Line2D.Float(0, 0.85f, 0, 0.7f));
          // Draw a N
          GeneralPath path = new GeneralPath();
          path.moveTo(-0.1f, -0.8f);
          path.lineTo(-0.1f, -1f);
          path.lineTo(0.1f, -0.8f);
          path.lineTo(0.1f, -1f);
          g2D.setStroke(new BasicStroke(1.5f / getWidth()));
          g2D.draw(path);
          // Draw the needle
          GeneralPath needlePath = new GeneralPath();
          needlePath.moveTo(0, -0.75f);
          needlePath.lineTo(0.2f, 0.7f);
          needlePath.lineTo(0, 0.5f);
          needlePath.lineTo(-0.2f, 0.7f);
          needlePath.closePath();
          needlePath.moveTo(-0.02f, 0);
          needlePath.lineTo(0.02f, 0);
          g2D.setStroke(new BasicStroke(4 / getWidth()));
          g2D.draw(needlePath);
        }
      };
    this.northDirectionComponent.setOpaque(false);
    northDirectionSpinnerModel.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent ev) {
        northDirection = (float)Math.toRadians(((Number)northDirectionSpinnerModel.getValue()).intValue());
        home.setVisualProperty(VISUAL_PROPERTY_PREFIX + "northDirection", northDirection);
        northDirectionComponent.repaint();
      }
    });
    
    this.latitudeLabel = new JLabel();
    final SpinnerNumberModel latitudeSpinnerModel = new SpinnerNumberModel(0., -90., 90., 10);
    this.latitudeSpinner = new JSpinner(latitudeSpinnerModel);
    // Change positive / negative notation by North / South
    JFormattedTextField textField = ((DefaultEditor)this.latitudeSpinner.getEditor()).getTextField();
    NumberFormatter numberFormatter = (NumberFormatter)((DefaultFormatterFactory)textField.getFormatterFactory()).getDefaultFormatter();
    numberFormatter.setFormat(new DecimalFormat("N ###.000;S ###.000"));
    textField.setFormatterFactory(new DefaultFormatterFactory(numberFormatter));
    SwingTools.addAutoSelectionOnFocusGain(textField);
    latitudeSpinnerModel.setValue(Math.toDegrees(this.latitude));
    latitudeSpinnerModel.addChangeListener(new ChangeListener() {
        public void stateChanged(ChangeEvent ev) {
          latitude = (float)Math.toRadians(((Number)latitudeSpinnerModel.getValue()).doubleValue());
          home.setVisualProperty(VISUAL_PROPERTY_PREFIX + "latitude", latitude);
        }
      });
    
    this.longitudeLabel = new JLabel();
    final SpinnerNumberModel longitudeSpinnerModel = new SpinnerNumberModel(0., -180., 180., 10);
    this.longitudeSpinner = new JSpinner(longitudeSpinnerModel);
    // Change positive / negative notation by East / West
    textField = ((DefaultEditor)this.longitudeSpinner.getEditor()).getTextField();
    numberFormatter = (NumberFormatter)((DefaultFormatterFactory)textField.getFormatterFactory()).getDefaultFormatter();
    numberFormatter.setFormat(new DecimalFormat("E ###.000;W ###.000"));
    textField.setFormatterFactory(new DefaultFormatterFactory(numberFormatter));
    SwingTools.addAutoSelectionOnFocusGain(textField);
    longitudeSpinnerModel.setValue(Math.toDegrees(this.longitude));
    longitudeSpinnerModel.addChangeListener(new ChangeListener() {
        public void stateChanged(ChangeEvent ev) {
          longitude = (float)Math.toRadians(((Number)longitudeSpinnerModel.getValue()).doubleValue());
          home.setVisualProperty(VISUAL_PROPERTY_PREFIX + "longitude", longitude);
        }
      });
    
    this.useCeilingLightCheckBox = new JCheckBox();
    this.useCeilingLightCheckBox.setSelected(this.useCeilingLight);
    this.useCeilingLightCheckBox.addItemListener(new ItemListener() {
        public void itemStateChanged(ItemEvent ev) {
          useCeilingLight = useCeilingLightCheckBox.isSelected();
          home.setVisualProperty(VISUAL_PROPERTY_PREFIX + "useCeilingLight", useCeilingLight); 
        }
      });

    this.useFurnitureLightCheckBox = new JCheckBox();
    this.useFurnitureLightCheckBox.setSelected(this.useFurnitureLight);
    this.useFurnitureLightCheckBox.addItemListener(new ItemListener() {
        public void itemStateChanged(ItemEvent ev) {
          useFurnitureLight = useFurnitureLightCheckBox.isSelected();
          home.setVisualProperty(VISUAL_PROPERTY_PREFIX + "useFurnitureLight", useFurnitureLight); 
        }
      });

    this.addSunlightCheckBox = new JCheckBox();
    this.addSunlightCheckBox.setSelected(this.addSunlight);
    this.addSunlightCheckBox.addItemListener(new ItemListener() {
        public void itemStateChanged(ItemEvent ev) {
          addSunlight = addSunlightCheckBox.isSelected();
          home.setVisualProperty(VISUAL_PROPERTY_PREFIX + "addSunLight", addSunlight); 
        }
      });
    this.shinyWallsCheckBox = new JCheckBox();
    this.shinyWallsCheckBox.setSelected(this.wallShininess > 1);
    this.shinyWallsCheckBox.addItemListener(new ItemListener() {
        public void itemStateChanged(ItemEvent ev) {
          wallShininess = shinyWallsCheckBox.isSelected() ? 30 : 0;
          home.setVisualProperty(VISUAL_PROPERTY_PREFIX + "wallShininess", wallShininess); 
        }
      });
    this.shinyFloorsCheckBox = new JCheckBox();
    this.shinyFloorsCheckBox.setSelected(this.floorShininess > 1);
    this.shinyFloorsCheckBox.addItemListener(new ItemListener() {
        public void itemStateChanged(ItemEvent ev) {
          floorShininess = shinyFloorsCheckBox.isSelected() ? 40 : 0;
          home.setVisualProperty(VISUAL_PROPERTY_PREFIX + "floorShininess", floorShininess); 
        }
      });
    this.shinyCeilingsCheckBox = new JCheckBox();
    this.shinyCeilingsCheckBox.setSelected(this.ceilingShininess > 1);
    this.shinyCeilingsCheckBox.addItemListener(new ItemListener() {
        public void itemStateChanged(ItemEvent ev) {
          ceilingShininess = shinyCeilingsCheckBox.isSelected() ? 30 : 0;
          home.setVisualProperty(VISUAL_PROPERTY_PREFIX + "ceilingShininess", ceilingShininess); 
        }
      });
    
    final ActionMap actionMap = getActionMap();
    this.createButton = new JButton(actionMap.get(ActionType.START_PHOTO_CREATION));
    this.createButton.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent ev) {
          // Swap Start / Stop action
          if (createButton.getAction() == actionMap.get(ActionType.START_PHOTO_CREATION)) {
            createButton.setAction(actionMap.get(ActionType.STOP_PHOTO_CREATION));
          } else {
            createButton.setAction(actionMap.get(ActionType.START_PHOTO_CREATION));
          }
        }
      });
    this.saveButton = new JButton(actionMap.get(ActionType.SAVE_PHOTO));
    this.closeButton = new JButton(actionMap.get(ActionType.CLOSE));

    setComponentTexts(preferences, additionalResource);
  }

  /**
   * Updates photo height.  
   */
  private void updatePhotoHeight() {
    if (this.photoAspectRatio != AspectRatio.FREE_RATIO) {
      this.heightSpinner.getModel().removeChangeListener(heightChangeListener);
      if (this.photoAspectRatio == AspectRatio.VIEW_3D_RATIO) {
        this.photoHeight = Math.round(this.photoWidth / this.view3DAspectRatio);
      } else if (this.photoAspectRatio == RATIO_2_1) {
        this.photoHeight = Math.round(photoWidth / 2);
      } else if (((AspectRatio)this.photoAspectRatio).getValue() != null) {
        this.photoHeight = Math.round(this.photoWidth / ((AspectRatio)this.photoAspectRatio).getValue());
      }
      this.heightSpinner.setValue(this.photoHeight);
      this.home.setVisualProperty(VISUAL_PROPERTY_PREFIX + "photoHeight", this.photoHeight);
      this.heightSpinner.getModel().addChangeListener(this.heightChangeListener);
    }
    boolean fixedProportions = this.lens == AdvancedPhotoRenderer.Lens.FISHEYE 
        || this.lens == AdvancedPhotoRenderer.Lens.SPHERICAL;
    this.applyProportionsCheckBox.setEnabled(!fixedProportions); 
    this.aspectRatioComboBox.setEnabled(!fixedProportions && applyProportionsCheckBox.isSelected()); 
  }
  
  /**
   * Sets the texts of the components.
   */
  private void setComponentTexts(UserPreferences preferences, ResourceBundle additionalResource) {
    this.widthLabel.setText(SwingTools.getLocalizedLabelText(preferences, 
        PhotoPanel.class, "widthLabel.text"));
    this.heightLabel.setText(SwingTools.getLocalizedLabelText(preferences, 
        PhotoPanel.class, "heightLabel.text"));
    this.applyProportionsCheckBox.setText(SwingTools.getLocalizedLabelText(preferences, 
        PhotoPanel.class, "applyProportionsCheckBox.text"));
    this.qualityLabel.setText(SwingTools.getLocalizedLabelText(preferences, 
        PhotoPanel.class, "qualityLabel.text"));
    this.fastQualityRadioButton.setText(SwingTools.getLocalizedLabelText(preferences,
        PhotoPanel.class, "fastLabel.text"));
    this.bestQualityRadioButton.setText(SwingTools.getLocalizedLabelText(preferences,
        PhotoPanel.class, "bestLabel.text"));
    
    this.lensLabel.setText(additionalResource.getString("AdvancedPhotoPanel.lensLabel.text")); 
    this.dateLabel.setText(additionalResource.getString("AdvancedPhotoPanel.dateLabel.text"));
    this.timeLabel.setText(additionalResource.getString("AdvancedPhotoPanel.timeLabel.text"));
    this.timeZoneLabel.setText(additionalResource.getString("AdvancedPhotoPanel.timeZoneLabel.text")); 
    this.northDirectionLabel.setText(additionalResource.getString("AdvancedPhotoPanel.northDirectionLabel.text")); 
    this.latitudeLabel.setText(additionalResource.getString("AdvancedPhotoPanel.latitudeLabel.text"));
    this.longitudeLabel.setText(additionalResource.getString("AdvancedPhotoPanel.longitudeLabel.text"));
    this.useCeilingLightCheckBox.setText(additionalResource.getString("AdvancedPhotoPanel.useCeilingLightCheckBox.text")); 
    this.useFurnitureLightCheckBox.setText(additionalResource.getString("AdvancedPhotoPanel.useFurnitureLightCheckBox.text")); 
    this.addSunlightCheckBox.setText(additionalResource.getString("AdvancedPhotoPanel.addSunlightCheckBox.text")); 
    this.shinyWallsCheckBox.setText(additionalResource.getString("AdvancedPhotoPanel.shinyWallsCheckBox.text")); 
    this.shinyFloorsCheckBox.setText(additionalResource.getString("AdvancedPhotoPanel.shinyFloorsCheckBox.text")); 
    this.shinyCeilingsCheckBox.setText(additionalResource.getString("AdvancedPhotoPanel.shinyCeilingsCheckBox.text")); 
    
    this.dialogTitle = additionalResource.getString("AdvancedPhotoPanel.title");
    Window window = SwingUtilities.getWindowAncestor(this);  
    if (window != null) {
      ((JDialog)window).setTitle(this.dialogTitle);
    }
    // Buttons text changes automatically through their action
  }

  /**
   * Sets components mnemonics and label / component associations.
   */
  private void setMnemonics(UserPreferences preferences, ResourceBundle additionalResource) {
    if (!OperatingSystem.isMacOSX()) {
      this.widthLabel.setDisplayedMnemonic(
          KeyStroke.getKeyStroke(preferences.getLocalizedString(
              PhotoPanel.class, "widthLabel.mnemonic")).getKeyCode());
      this.widthLabel.setLabelFor(this.widthSpinner);
      this.heightLabel.setDisplayedMnemonic(
          KeyStroke.getKeyStroke(preferences.getLocalizedString(
              PhotoPanel.class, "heightLabel.mnemonic")).getKeyCode());
      this.heightLabel.setLabelFor(this.heightSpinner);
      this.applyProportionsCheckBox.setMnemonic(
          KeyStroke.getKeyStroke(preferences.getLocalizedString(
              PhotoPanel.class, "applyProportionsCheckBox.mnemonic")).getKeyCode());
      
      this.lensLabel.setDisplayedMnemonic(KeyStroke.getKeyStroke(
          additionalResource.getString("AdvancedPhotoPanel.lensLabel.mnemonic")).getKeyCode());
      this.lensLabel.setLabelFor(this.lensComboBox);
      this.dateLabel.setDisplayedMnemonic(KeyStroke.getKeyStroke(
          additionalResource.getString("AdvancedPhotoPanel.dateLabel.mnemonic")).getKeyCode());
      this.dateLabel.setLabelFor(this.dateSpinner);
      this.timeLabel.setDisplayedMnemonic(KeyStroke.getKeyStroke(
          additionalResource.getString("AdvancedPhotoPanel.timeLabel.mnemonic")).getKeyCode());
      this.timeLabel.setLabelFor(this.timeSpinner);
      this.timeZoneLabel.setDisplayedMnemonic(KeyStroke.getKeyStroke(
          additionalResource.getString("AdvancedPhotoPanel.timeZoneLabel.mnemonic")).getKeyCode());
      this.timeZoneLabel.setLabelFor(this.timeZoneComboBox);
      this.northDirectionLabel.setDisplayedMnemonic(KeyStroke.getKeyStroke(
          additionalResource.getString("AdvancedPhotoPanel.northDirectionLabel.mnemonic")).getKeyCode());
      this.northDirectionLabel.setLabelFor(this.northDirectionSpinner);
      this.latitudeLabel.setDisplayedMnemonic(KeyStroke.getKeyStroke(
          additionalResource.getString("AdvancedPhotoPanel.latitudeLabel.mnemonic")).getKeyCode());
      this.latitudeLabel.setLabelFor(this.latitudeSpinner);
      this.longitudeLabel.setDisplayedMnemonic(KeyStroke.getKeyStroke(
          additionalResource.getString("AdvancedPhotoPanel.longitudeLabel.mnemonic")).getKeyCode());
      this.longitudeLabel.setLabelFor(this.longitudeSpinner);
      this.useCeilingLightCheckBox.setMnemonic(KeyStroke.getKeyStroke(
          additionalResource.getString("AdvancedPhotoPanel.useCeilingLightCheckBox.mnemonic")).getKeyCode());
      this.useFurnitureLightCheckBox.setMnemonic(KeyStroke.getKeyStroke(
          additionalResource.getString("AdvancedPhotoPanel.useFurnitureLightCheckBox.mnemonic")).getKeyCode());
      this.addSunlightCheckBox.setMnemonic(KeyStroke.getKeyStroke(
          additionalResource.getString("AdvancedPhotoPanel.addSunlightCheckBox.mnemonic")).getKeyCode());
      this.shinyWallsCheckBox.setMnemonic(KeyStroke.getKeyStroke(
          additionalResource.getString("AdvancedPhotoPanel.shinyWallsCheckBox.mnemonic")).getKeyCode());
      this.shinyFloorsCheckBox.setMnemonic(KeyStroke.getKeyStroke(
          additionalResource.getString("AdvancedPhotoPanel.shinyFloorsCheckBox.mnemonic")).getKeyCode());
      this.shinyCeilingsCheckBox.setMnemonic(KeyStroke.getKeyStroke(
          additionalResource.getString("AdvancedPhotoPanel.shinyCeilingsCheckBox.mnemonic")).getKeyCode());
    }
  }

  /**
   * Preferences property listener bound to this panel with a weak reference to avoid
   * strong link between user preferences and this panel.  
   */
  public static class LanguageChangeListener implements PropertyChangeListener {
    private final WeakReference<AdvancedPhotoPanel> photoPanel;
    private final ResourceBundle                    additionalResource;

    public LanguageChangeListener(AdvancedPhotoPanel photoPanel, 
                                  ResourceBundle additionalResource) {
      this.photoPanel = new WeakReference<AdvancedPhotoPanel>(photoPanel);
      this.additionalResource = additionalResource;
    }

    public void propertyChange(PropertyChangeEvent ev) {
      // If photo panel was garbage collected, remove this listener from preferences
      AdvancedPhotoPanel photoPanel = this.photoPanel.get();
      UserPreferences preferences = (UserPreferences)ev.getSource();
      if (photoPanel == null) {
        preferences.removePropertyChangeListener(UserPreferences.Property.LANGUAGE, this);
      } else {
        photoPanel.setComponentTexts(preferences, this.additionalResource);
        photoPanel.setMnemonics(preferences, this.additionalResource);
      }
    }
  }

  /**
   * Layouts panel components in panel with their labels. 
   */
  private void layoutComponents() {
    int labelAlignment = OperatingSystem.isMacOSX() 
        ? GridBagConstraints.LINE_END
        : GridBagConstraints.LINE_START;
    // Add animatedWaitLabel and photoComponent to a card panel 
    this.photoCardLayout = new CardLayout();
    this.photoPanel = new JPanel(this.photoCardLayout);
    photoPanel.add(this.photoComponent, PHOTO_CARD);
    photoPanel.add(this.animatedWaitLabel, WAIT_CARD);
    int gapHeight = 5;
    // First row
    Insets lastComponentInsets = new Insets(0, 0, gapHeight, 0);
    add(photoPanel, new GridBagConstraints(
        0, 0, 6, 1, 1, 1, labelAlignment, 
        GridBagConstraints.BOTH, lastComponentInsets, 0, 0));
    // Second row
    // Add a dummy label at left and right
    add(new JLabel(), new GridBagConstraints(
        0, 1, 1, 3, 0.5f, 0, GridBagConstraints.CENTER, 
        GridBagConstraints.BOTH, new Insets(0, 0, 0, 0), 0, 0));
    add(new JLabel(), new GridBagConstraints(
        5, 1, 1, 3, 0.5f, 0, GridBagConstraints.CENTER, 
        GridBagConstraints.BOTH, new Insets(0, 0, 0, 0), 0, 0));
    Insets labelInsets = new Insets(0, 0, gapHeight, 5);
    Insets componentInsets = new Insets(0, 0, gapHeight, 10);
    add(this.widthLabel, new GridBagConstraints(
        1, 1, 1, 1, 0, 0, labelAlignment, 
        GridBagConstraints.NONE, labelInsets, 0, 0));
    add(this.widthSpinner, new GridBagConstraints(
        2, 1, 1, 1, 0, 0, GridBagConstraints.LINE_START, 
        GridBagConstraints.HORIZONTAL, componentInsets, 0, 0));
    add(this.heightLabel, new GridBagConstraints(
        3, 1, 1, 1, 0, 0, labelAlignment, 
        GridBagConstraints.NONE, labelInsets, 0, 0));
    add(this.heightSpinner, new GridBagConstraints(
        4, 1, 1, 1, 0, 0, GridBagConstraints.LINE_START, 
        GridBagConstraints.HORIZONTAL, lastComponentInsets, 0, 0));
    // Third and fourth row
    JPanel proportionsAndQualityPanel = new JPanel(new GridBagLayout());
    proportionsAndQualityPanel.add(this.applyProportionsCheckBox, new GridBagConstraints(
        0, 0, 1, 1, 0, 0, labelAlignment, 
        GridBagConstraints.NONE, OperatingSystem.isMacOSX() ? new Insets(0, 0, gapHeight, 0) : labelInsets, 0, 0));
    proportionsAndQualityPanel.add(this.aspectRatioComboBox, new GridBagConstraints(
        1, 0, 1, 1, 0, 0, GridBagConstraints.LINE_START,
        GridBagConstraints.HORIZONTAL, new Insets(0, 0, gapHeight, 10), 0, 0));
    proportionsAndQualityPanel.add(this.qualityLabel, new GridBagConstraints(
        2, 0, 1, 1, 0, 0, labelAlignment, 
        GridBagConstraints.NONE, labelInsets, 0, 0));
    proportionsAndQualityPanel.add(this.fastQualityRadioButton, new GridBagConstraints(
        3, 0, 1, 1, 0, 0, GridBagConstraints.LINE_START, 
        GridBagConstraints.NONE, lastComponentInsets, 0, 0));
    
    proportionsAndQualityPanel.add(this.lensLabel, new GridBagConstraints(
        0, 1, 1, 1, 0, 0, labelAlignment, 
        GridBagConstraints.NONE, labelInsets, 0, 0));    
    this.lensComboBox.setPreferredSize(this.aspectRatioComboBox.getPreferredSize());
    proportionsAndQualityPanel.add(this.lensComboBox, new GridBagConstraints(
        1, 1, 1, 1, 0, 0, GridBagConstraints.LINE_START, 
        GridBagConstraints.HORIZONTAL, new Insets(0, 0, gapHeight, 10), 0, 0));    
    proportionsAndQualityPanel.add(this.bestQualityRadioButton, new GridBagConstraints(
        3, 1, 1, 1, 0, 0, GridBagConstraints.LINE_START, 
        GridBagConstraints.NONE, lastComponentInsets, 0, 0));    

    add(proportionsAndQualityPanel, new GridBagConstraints(
        1, 2, 4, 1, 0, 0, GridBagConstraints.CENTER, 
        GridBagConstraints.NONE, new Insets(0, 0, 0, 0), 0, 0));
    // Fifth row
    add(new JSeparator(), new GridBagConstraints(
        1, 4, 4, 1, 0, 0, GridBagConstraints.CENTER, 
        GridBagConstraints.HORIZONTAL, lastComponentInsets, 0, 0));
    // Sixth row
    add(this.dateLabel, new GridBagConstraints(
        1, 5, 1, 1, 0, 0, labelAlignment, 
        GridBagConstraints.NONE, labelInsets, 0, 0));
    add(this.dateSpinner, new GridBagConstraints(
        2, 5, 1, 1, 0, 0, GridBagConstraints.LINE_START, 
        GridBagConstraints.HORIZONTAL, componentInsets, 0, 0));
    add(this.timeLabel, new GridBagConstraints(
        3, 5, 1, 1, 0, 0, labelAlignment, 
        GridBagConstraints.NONE, labelInsets, 0, 0));
    add(this.timeSpinner, new GridBagConstraints(
        4, 5, 1, 1, 0, 0, GridBagConstraints.LINE_START, 
        GridBagConstraints.HORIZONTAL, lastComponentInsets, 0, 0));
    // Seventh row
    add(this.timeZoneLabel, new GridBagConstraints(
        1, 6, 1, 1, 0, 0, labelAlignment, 
        GridBagConstraints.NONE, labelInsets, 0, 0));
    add(this.timeZoneComboBox, new GridBagConstraints(
        2, 6, 1, 1, 0, 0, GridBagConstraints.LINE_START, 
        GridBagConstraints.HORIZONTAL, componentInsets, 30, 0));
    this.timeZoneComboBox.setPreferredSize(this.dateSpinner.getPreferredSize());
    add(this.northDirectionLabel, new GridBagConstraints(
        3, 6, 1, 1, 0, 0, labelAlignment, 
        GridBagConstraints.NONE, labelInsets, 0, 0));
    JPanel northDirectionPanel = new JPanel(new FlowLayout(FlowLayout.LEFT, 0, 0));
    northDirectionPanel.add(this.northDirectionSpinner);
    northDirectionPanel.add(new JLabel(" "));
    northDirectionPanel.add(this.northDirectionComponent);
    add(northDirectionPanel, new GridBagConstraints(
        4, 6, 1, 1, 0, 0, GridBagConstraints.LINE_START, 
        GridBagConstraints.HORIZONTAL, lastComponentInsets, 0, 0));
    // Height row
    add(this.latitudeLabel, new GridBagConstraints(
        1, 7, 1, 1, 0, 0, labelAlignment, 
        GridBagConstraints.NONE, labelInsets, 0, 0));
    add(this.latitudeSpinner, new GridBagConstraints(
        2, 7, 1, 1, 0, 0, GridBagConstraints.LINE_START, 
        GridBagConstraints.HORIZONTAL, componentInsets, 0, 0));
    add(this.longitudeLabel, new GridBagConstraints(
        3, 7, 1, 1, 0, 0, labelAlignment, 
        GridBagConstraints.NONE, labelInsets, 0, 0));
    add(this.longitudeSpinner, new GridBagConstraints(
        4, 7, 1, 1, 0, 0, GridBagConstraints.LINE_START, 
        GridBagConstraints.HORIZONTAL, lastComponentInsets, 0, 0));
    // Ninth row
    JPanel lightsPanel = new JPanel(new GridBagLayout());
    lightsPanel.add(this.useCeilingLightCheckBox, new GridBagConstraints(
        0, 0, 1, 1, 0, 0, GridBagConstraints.WEST, 
        GridBagConstraints.NONE, new Insets(0, 0, gapHeight, 2), 0, 0));
    lightsPanel.add(this.useFurnitureLightCheckBox, new GridBagConstraints(
        1, 0, 1, 1, 0, 0, GridBagConstraints.WEST, 
        GridBagConstraints.NONE, new Insets(0, 0, gapHeight, 2), 0, 0));
    lightsPanel.add(this.addSunlightCheckBox, new GridBagConstraints(
        2, 0, 1, 1, 0, 0, GridBagConstraints.WEST, 
        GridBagConstraints.NONE, lastComponentInsets, 0, 0));
    lightsPanel.add(this.shinyWallsCheckBox, new GridBagConstraints(
        0, 1, 1, 1, 0, 0, GridBagConstraints.WEST, 
        GridBagConstraints.NONE, new Insets(0, 0, 0, 2), 0, 0));
    lightsPanel.add(this.shinyFloorsCheckBox, new GridBagConstraints(
        1, 1, 1, 1, 0, 0, GridBagConstraints.WEST, 
        GridBagConstraints.NONE, new Insets(0, 0, 0, 2), 0, 0));
    lightsPanel.add(this.shinyCeilingsCheckBox, new GridBagConstraints(
        2, 1, 1, 1, 0, 0, GridBagConstraints.WEST, 
        GridBagConstraints.NONE, new Insets(0, 0, 0, 0), 0, 0));
    add(lightsPanel, new GridBagConstraints(
        1, 8, 4, 1, 0, 0, GridBagConstraints.CENTER, 
        GridBagConstraints.NONE, new Insets(0, 0, 0, 0), 0, 0));
  }
  
  /**
   * Displays this panel in a non modal dialog.
   */
  public void displayView(Container parent) {
    if (currentPhotoPanel == this) {
      SwingUtilities.getWindowAncestor(AdvancedPhotoPanel.this).toFront();
    } else {
      if (currentPhotoPanel != null) {
        currentPhotoPanel.close();
      }
      final JOptionPane optionPane = new JOptionPane(this, 
          JOptionPane.PLAIN_MESSAGE, JOptionPane.OK_CANCEL_OPTION,
          null, new Object [] {this.createButton, this.saveButton, this.closeButton}, this.createButton);
      final JDialog dialog = optionPane.createDialog(parent, this.dialogTitle);
      dialog.setModal(false);
      dialog.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
      dialog.addComponentListener(new ComponentAdapter() {
          @Override
          public void componentHidden(ComponentEvent ev) {
            if (optionPane.getValue() != null
                && optionPane.getValue() != JOptionPane.UNINITIALIZED_VALUE) {
              close();
            }
          }
        });
      
      dialog.setLocationByPlatform(true);
      Component homeRoot = SwingUtilities.getRoot(parent);
      if (parent.getParent() instanceof Window) {
        // Get root's root
        homeRoot = SwingUtilities.getRoot(homeRoot);
      }
      if (homeRoot != null) {
        int windowRightBorder = homeRoot.getX() + homeRoot.getWidth();
        Dimension screenSize = getToolkit().getScreenSize();
        Insets screenInsets = getToolkit().getScreenInsets(getGraphicsConfiguration());
        int screenRightBorder = screenSize.width - screenInsets.right;
        int dialogWidth = dialog.getWidth();
        // If there some space left at the right of the window
        if (screenRightBorder - windowRightBorder > dialogWidth / 2) {
          // Move the dialog to the right of window
          dialog.setLocationByPlatform(false);
          dialog.setLocation(Math.min(windowRightBorder + 5, screenRightBorder - dialogWidth), 
              Math.max(Math.min(homeRoot.getY() + dialog.getInsets().top * 2, 
                  screenSize.height - dialog.getHeight() - screenInsets.bottom), screenInsets.top));
        }
      }
      
      final JComponent view3D = (JComponent)findChild((Container)homeRoot, HomeComponent3D.class);
      // Add a listener on 3D view to be notified when its size changes 
      final ComponentAdapter view3DSizeListener = new ComponentAdapter() {
          @Override
          public void componentResized(ComponentEvent ev) {
            view3DAspectRatio = (float)view3D.getWidth() / view3D.getHeight();
            updatePhotoHeight();
          }
        };
      if (view3D != null) {
        this.view3DAspectRatio = (float)view3D.getWidth() / view3D.getHeight();
        updatePhotoHeight();
        view3D.addComponentListener(view3DSizeListener);
      }
      dialog.addWindowListener(new WindowAdapter() {
          public void windowClosed(WindowEvent ev) {
            if (view3D != null) {
              view3D.removeComponentListener(view3DSizeListener);
            }
            stopPhotoCreation();
            currentPhotoPanel = null;
          }
        });

      dialog.setVisible(true);
      currentPhotoPanel = this;
    }
  }

  /**
   * Returns the first child of <code>parent</code> of the given type.
   */
  private Component findChild(Container parent, Class<?> childClass) {
    for (int i = 0; i < parent.getComponentCount(); i++) {
      Component child = parent.getComponent(i);
      if (childClass.isInstance(child)) {
        return child;
      } else if (child instanceof Container) {
        child = findChild((Container)child, childClass);
        if (child != null) {
          return child;
        }
      }            
    }
    return null;
  }
  
  /**
   * Creates the photo image depending on the quality requested by the user.
   */
  private void startPhotoCreation() {
    this.photoComponent.setImage(null);
    this.widthSpinner.setEnabled(false);
    this.heightSpinner.setEnabled(false);
    this.lensComboBox.setEnabled(false);
    this.applyProportionsCheckBox.setEnabled(false);
    this.aspectRatioComboBox.setEnabled(false);
    this.fastQualityRadioButton.setEnabled(false);
    this.bestQualityRadioButton.setEnabled(false);
    this.dateSpinner.setEnabled(false);
    this.timeSpinner.setEnabled(false);
    this.timeZoneComboBox.setEnabled(false);
    this.northDirectionSpinner.setEnabled(false);
    this.longitudeSpinner.setEnabled(false);
    this.latitudeSpinner.setEnabled(false);
    this.useCeilingLightCheckBox.setEnabled(false);
    this.useFurnitureLightCheckBox.setEnabled(false);
    this.addSunlightCheckBox.setEnabled(false);
    this.shinyWallsCheckBox.setEnabled(false);
    this.shinyFloorsCheckBox.setEnabled(false);
    this.shinyCeilingsCheckBox.setEnabled(false);
    getActionMap().get(ActionType.SAVE_PHOTO).setEnabled(false);
    this.photoCardLayout.show(this.photoPanel, WAIT_CARD);
    
    // Compute photo in an other executor thread
    // Use a clone of home because the user can modify home during photo computation
    final Home home = this.home.clone();
    this.photoCreationExecutor = Executors.newSingleThreadExecutor();
    this.photoCreationExecutor.execute(new Runnable() {
        public void run() {
          computePhoto(home);
        }
      });
  }

  /**
   * Computes the photo of the given home.
   * Caution : this method must be thread safe because it's called from an executor. 
   */
  private void computePhoto(Home home) {
    BufferedImage image = null;
    try {
      int quality = this.photoQuality;
      int imageWidth = this.photoWidth;
      int imageHeight = this.photoHeight;
      // Use advanced photo renderer
      AdvancedPhotoRenderer photoRenderer = new AdvancedPhotoRenderer(home, 
          quality == 2 
              ? AdvancedPhotoRenderer.Quality.LOW 
              : AdvancedPhotoRenderer.Quality.HIGH,
          this.lens,
          this.photoCalendar,
          this.northDirection,
          this.latitude,
          this.longitude,
          true,
          this.useCeilingLight,
          this.useFurnitureLight,
          this.addSunlight,
          this.wallShininess,
          this.floorShininess,
          this.ceilingShininess);
      if (!Thread.currentThread().isInterrupted()) {
        image = new BufferedImage(imageWidth, imageHeight, BufferedImage.TYPE_INT_RGB);
        this.photoComponent.setImage(image);
        EventQueue.invokeLater(new Runnable() {
          public void run() {
            photoCardLayout.show(photoPanel, PHOTO_CARD);
          }
        });
        photoRenderer.render(image, home.getCamera(), this.photoComponent);
      }
    } catch (OutOfMemoryError ex) {
      image = getErrorImage();
      throw ex;
    } catch (IllegalStateException ex) {
      image = getErrorImage();
      throw ex;
    } catch (IOException ex) {
      image = getErrorImage();
    } finally {           
      final BufferedImage photoImage = this.photoCreationExecutor != null
          ? image
          : null;
      EventQueue.invokeLater(new Runnable() {
          public void run() {
            getActionMap().get(ActionType.SAVE_PHOTO).setEnabled(photoImage != null);
            createButton.setAction(getActionMap().get(ActionType.START_PHOTO_CREATION));
            photoComponent.setImage(photoImage);
            widthSpinner.setEnabled(true);
            heightSpinner.setEnabled(true);
            lensComboBox.setEnabled(true);
            boolean fixedProportions = lens == AdvancedPhotoRenderer.Lens.FISHEYE 
                || lens == AdvancedPhotoRenderer.Lens.SPHERICAL;
            applyProportionsCheckBox.setEnabled(!fixedProportions); 
            aspectRatioComboBox.setEnabled(!fixedProportions && applyProportionsCheckBox.isSelected());
            fastQualityRadioButton.setEnabled(true);
            bestQualityRadioButton.setEnabled(true);
            dateSpinner.setEnabled(true);
            timeSpinner.setEnabled(true);
            timeZoneComboBox.setEnabled(true);
            northDirectionSpinner.setEnabled(true);
            longitudeSpinner.setEnabled(true);
            latitudeSpinner.setEnabled(true);
            useCeilingLightCheckBox.setEnabled(true);
            useFurnitureLightCheckBox.setEnabled(true);
            addSunlightCheckBox.setEnabled(true);
            shinyWallsCheckBox.setEnabled(true);
            shinyFloorsCheckBox.setEnabled(true);
            shinyCeilingsCheckBox.setEnabled(true);
            photoCardLayout.show(photoPanel, PHOTO_CARD);
            photoCreationExecutor = null;
          }
        });
    }
  }
  
  /**
   * Returns the image used in case of an error.
   */
  private BufferedImage getErrorImage() {
    Icon errorIcon = IconManager.getInstance().getErrorIcon(16);
    BufferedImage errorImage = new BufferedImage(
        errorIcon.getIconWidth(), errorIcon.getIconHeight(), BufferedImage.TYPE_INT_ARGB);
    Graphics2D g2D = (Graphics2D)errorImage.getGraphics();
    errorIcon.paintIcon(this, g2D, 0, 0);
    g2D.dispose();
    return errorImage;
  }
  
  /**
   * Stops photo creation.
   */
  private void stopPhotoCreation() {
    if (this.photoCreationExecutor != null) {
      // Will interrupt executor thread
      this.photoCreationExecutor.shutdownNow();
      this.photoCreationExecutor = null;
    }
  }

  /**
   * Saves the created image.
   */
  private void savePhoto() {
    String pngFile = new FileContentManager(preferences).showSaveDialog((View)findChild(getRootPane(), View.class),
        this.preferences.getLocalizedString(PhotoPanel.class, "savePhotoDialog.title"), 
        FileContentManager.ContentType.PNG, this.home.getName());
    try {
      if (pngFile != null) {
        ImageIO.write(this.photoComponent.getImage(), "PNG", new File(pngFile));
      }
    } catch (IOException ex) {
      String messageFormat = this.preferences.getLocalizedString(PhotoPanel.class, "savePhotoError.message");
      JOptionPane.showMessageDialog(SwingUtilities.getRootPane(this), String.format(messageFormat, ex.getMessage()), 
          this.preferences.getLocalizedString(PhotoPanel.class, "savePhotoError.title"), JOptionPane.ERROR_MESSAGE);
    }
  }

  /**
   * Manages closing of this pane.
   */
  private void close() {
    Window window = SwingUtilities.getWindowAncestor(this);
    if (window.isDisplayable()) {
      window.dispose();
    }    
  }
}
