/*
 * CreateAdvancedPhotoPlugin.java  
 *
 * Copyright (c) 2010 Emmanuel PUYBARET / eTeks <info@eteks.com>. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.eteks.sweethome3d.plugin.advancedrendering;

import java.awt.KeyboardFocusManager;
import java.awt.Window;
import java.util.Locale;
import java.util.ResourceBundle;

import javax.jnlp.BasicService;
import javax.jnlp.ServiceManager;
import javax.jnlp.UnavailableServiceException;
import javax.swing.JEditorPane;
import javax.swing.JOptionPane;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;

import com.eteks.sweethome3d.plugin.Plugin;
import com.eteks.sweethome3d.plugin.PluginAction;

/**
 * A plug-in to create photo with advanced parameters.
 * With the contribution of <a href="http://www.sweethome3d.eu/support/forum/viewthread_thread,1370">Frdric Mantegazza</a>. 
 * @author Emmanuel Puybaret
 */
public class CreateAdvancedPhotoPlugin extends Plugin {
  @Override
  public PluginAction [] getActions() {
    return new PluginAction [] {new PluginAction(
            "com.eteks.sweethome3d.plugin.advancedrendering.ApplicationPlugin", 
            "CREATE_ADVANCED_PHOTO", getPluginClassLoader(), true) {
        @Override
        public void execute() {
          createAdvancedPhotoPlugin();
        }
      }
    };
  }
  
  /**
   * Displays and creates a dialog to create photos with advanced parameters.
   */
  private void createAdvancedPhotoPlugin() {
    final ResourceBundle resource = ResourceBundle.getBundle(
        "com.eteks.sweethome3d.plugin.advancedrendering.ApplicationPlugin", 
        Locale.getDefault(), getPluginClassLoader());
    final Window parent = KeyboardFocusManager.getCurrentKeyboardFocusManager().getActiveWindow();
    
    try {
      // Ignore plug-in in protected Java Web Start environment 
      ServiceManager.lookup("javax.jnlp.FileSaveService");
      // Use an uneditable editor pane to let user select text in dialog
      JEditorPane messagePane = new JEditorPane("text/html", 
          resource.getString("createAdvancedPhotoJavaWebStartInfo.message"));
      messagePane.setOpaque(false);
      messagePane.setEditable(false);
      try { 
        // Lookup the javax.jnlp.BasicService object 
        final BasicService service = (BasicService)ServiceManager.lookup("javax.jnlp.BasicService");
        // If basic service supports  web browser
        if (service.isWebBrowserSupported()) {
          // Add a listener that displays hyperlinks content in browser
          messagePane.addHyperlinkListener(new HyperlinkListener() {
            public void hyperlinkUpdate(HyperlinkEvent ev) {
              if (ev.getEventType() == HyperlinkEvent.EventType.ACTIVATED) {
                service.showDocument(ev.getURL()); 
              }
            }
          });
        }
      } catch (UnavailableServiceException ex) {
        // Too bad : service is unavailable             
      } 
      
      String title = resource.getString("createAdvancedPhotoJavaWebStartInfo.title");
      JOptionPane.showMessageDialog(parent, messagePane, title, JOptionPane.WARNING_MESSAGE);
      return;
    } catch (UnavailableServiceException ex) {
    }

    new AdvancedPhotoPanel(getHome(), getUserPreferences(), resource).displayView(parent);
  }
}
